# The render process for the cli does NOT use the rend project for a couple
# of reasons
# 1. Config needs to be very early in the startup, therefore it cannot use
#    asyncio, all rend funcs use asyncio
# 2. Config will not allow for template wrapping as the render is just a
#    single command line render


def process(hub, renderer, value):
    """
    Take a renderer and a value, process it, and return the processed value

    This is intended to load a string through the config render system
    """
    return getattr(hub, f"config.render.{renderer}.render")(value)


def pipe(hub, dpipe, data):
    """
    Given a render pipe, render the given data
    """
    for render in dpipe:
        if isinstance(render, bytes):
            render = render.decode()
        data = hub.config.render.init.process(render, data)
    return data


def load_file(hub, renderer, fn):
    """
    Load up a file with the passed renderer unless the file contains a
    renderer she-bang line
    """
    with open(fn, "rb") as rfh:
        data = rfh.read()
    if not data:
        return {}
    if data.startswith(b"#!"):
        dpipe = data[2 : data.index(b"\n")].split(b"|")
        return hub.config.render.init.pipe(dpipe, data)
    else:
        return hub.config.render.init.process(renderer, data)



"""
This sub is used to set up logging for pop projects and injects logging
options into conf making it easy to add robust logging
"""
# Import python libs
import logging


def __init__(hub):
    """
    Set up variables used by the log subsystem
    """
    logging.addLevelName(5, "TRACE")
    hub.log.LOGGER = {
        # Initialize the root logger
        "": logging.getLogger()
    }
    hub.log.LEVEL = {
        "notset": logging.NOTSET,
        "trace": 5,
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warn": logging.WARN,
        "warning": logging.WARNING,
        "error": logging.ERROR,
        "fatal": logging.FATAL,
        "critical": logging.CRITICAL,
    }

    # These should be overwritten by the integrated logger, but here's a contingency
    hub.log.INT_LEVEL = logging.INFO
    hub.log.log = hub.log.LOGGER[""].log
    hub.log.trace = lambda *args, **kwargs: hub.log.LOGGER[""].log(5, *args, **kwargs)
    hub.log.debug = hub.log.LOGGER[""].debug
    hub.log.info = hub.log.LOGGER[""].info
    hub.log.critical = hub.log.LOGGER[""].critical
    hub.log.warning = hub.log.LOGGER[""].warning
    hub.log.error = hub.log.LOGGER[""].error


def conf(hub, name):
    """
    Return the conf dict for logging, this should be merged OVER by the loaded
    config dict(s)
    """
    # TODO: Make this more robust to handle more logging interfaces
    ldict = {
        "log_file": {
            "default": f"{name}.log",
            "help": "The location of the log file",
            "group": "Logging Options",
        },
        "log_level": {
            "default": "info",
            "help": "Set the log level, either quiet, info, warning, or error",
            "group": "Logging Options",
        },
        "log_fmt_logfile": {
            "default": "%(asctime)s,%(msecs)03d [%(name)-17s][%(levelname)-8s] %(message)s",
            "help": "The format to be given to log file messages",
            "group": "Logging Options",
        },
        "log_fmt_console": {
            "default": "[%(levelname)-8s] %(message)s",
            "help": "The log formatting used in the console",
            "group": "Logging Options",
        },
        "log_datefmt": {
            "default": "%H:%M:%S",
            "help": "The date format to display in the logs",
            "group": "Logging Options",
        },
        "log_plugin": {
            "default": "basic",
            "help": "The logging plugin to use",
            "group": "Logging Options",
        },
    }
    return ldict



from typing import List


def load(
    hub,
    sources: List[str],
    cli: str = None,
    dyne_names: List[str] = None,
    loader: str = "yaml",
    parse_cli: bool = True,
    logs: bool = True,
):
    """
    Load up the configs from the integrate system
    """
    if not isinstance(sources, list):
        sources = [sources]
    if dyne_names is None:
        dyne_names = []
    raw = hub.config.dirs.load(sources, dyne_names, cli)
    os_vars = hub.config.os.init.gather(raw)
    cli_args, raw_cli = hub.config.args.gather(raw, cli, parse_cli)
    if cli_args.get("version"):
        hub.config.version.run(cli)
    configs = hub.config.file.init.parse(raw, cli, os_vars, cli_args, loader)
    opt = hub.config.order.apply(raw, raw_cli, cli, cli_args, os_vars, configs)
    hub.OPT = hub.pop.data.imap(opt)

    if logs:
        log_plugin = hub.OPT[sources[0]].get("log_plugin")
        getattr(hub, f"log.{log_plugin}.setup")(hub.OPT[sources[0]])



import inspect
import logging
import pop.contract
import sys


def _stack_frames(relative_start: int) -> inspect.FrameInfo:
    """
    Efficiently access stack frames.
    :param relative_start: Starting stack depth; The default, 2 is the parent of the
                           caller of stack_frames - the first function that may be unknown.
    :return: a stack frame
    """
    if hasattr(sys, "_getframe"):
        # implementation detail of CPython, speeds things up by 100x.
        frame = sys._getframe(relative_start)
        while frame:
            yield frame
            frame = frame.f_back
    else:
        for frame_info in inspect.stack(context=0)[relative_start:]:
            yield frame_info.frame


def _get_hub_ref() -> str:
    # Minimize lookup time by starting at frame 5, it will be at least that far back
    for frame in _stack_frames(5):
        if isinstance(frame.f_locals.get("self"), pop.contract.Contracted):
            contracted = frame.f_locals["self"]
            break
    else:
        # Default to the root reference
        return "hub"

    return contracted.func.__module__


def _get_logger(hub, name: str = "") -> logging.Logger:
    if name not in hub.log.LOGGER:
        hub.log.LOGGER[name] = logging.getLogger(name)
    return hub.log.LOGGER[name]


def log(hub, level: int, msg: str, *args, **kwargs):
    if hub.log.INT_LEVEL <= level:
        caller = _get_hub_ref()
        logger = _get_logger(hub, caller)
        logger.log(level, msg, *args, **kwargs)


def trace(hub, msg: str, *args, **kwargs):
    caller = _get_hub_ref()
    logger = _get_logger(hub, caller)
    logger.log(5, msg, *args, **kwargs)


def debug(hub, msg: str, *args, **kwargs):
    caller = _get_hub_ref()
    logger = _get_logger(hub, caller)
    logger.debug(msg, *args, **kwargs)


def info(hub, msg: str, *args, **kwargs):
    caller = _get_hub_ref()
    logger = _get_logger(hub, caller)
    logger.info(msg, *args, **kwargs)


def critical(hub, msg: str, *args, **kwargs):
    caller = _get_hub_ref()
    logger = _get_logger(hub, caller)
    logger.critical(msg, *args, **kwargs)


def warning(hub, msg: str, *args, **kwargs):
    caller = _get_hub_ref()
    logger = _get_logger(hub, caller)
    logger.warning(msg, *args, **kwargs)


def error(hub, msg: str, *args, **kwargs):
    caller = _get_hub_ref()
    logger = _get_logger(hub, caller)
    logger.error(msg, *args, **kwargs)


def null(hub, *args, **kwargs):
    """
    A logging level that eats log messages, this is because the functions defined above
    call `_get_hub_ref` which cat be expensive, especially if `sys` doesn't
    have the `_get_frame` function.
    """


def setup(hub, conf):
    """
    Given the configuration data set up the logger
    """
    raw_level = conf["log_level"].strip().lower()
    if raw_level.isdigit():
        hub.log.INT_LEVEL = int(raw_level)
    else:
        hub.log.INT_LEVEL = hub.log.LEVEL.get(raw_level, logging.DEBUG)

    # Use the saved root logger
    root = _get_logger(hub)
    root.setLevel(hub.log.INT_LEVEL)
    cf = logging.Formatter(fmt=conf["log_fmt_console"], datefmt=conf["log_datefmt"])
    ch = logging.StreamHandler()
    ch.setLevel(hub.log.INT_LEVEL)
    ch.setFormatter(cf)
    root.addHandler(ch)
    ff = logging.Formatter(fmt=conf["log_fmt_console"], datefmt=conf["log_datefmt"])
    fh = logging.FileHandler(conf["log_file"])
    fh.setLevel(hub.log.INT_LEVEL)
    fh.setFormatter(ff)
    root.addHandler(fh)

    hub.log.log = hub.log.basic.log
    for log_level in ("trace", "debug", "info", "critical", "warning", "error"):
        if hub.log.INT_LEVEL <= hub.log.LEVEL[log_level]:
            hub.log.basic.debug(f"Setting up {log_level} logging")
            # Set up logging at a low level on the hub
            setattr(hub, f"log.{log_level}", getattr(hub, f"log.basic.{log_level}"))
        else:
            hub.log.basic.debug(f"Ignoring {log_level} logging")
            # Override the logger with the null logger if we aren't at that level.
            # This significantly reduces unnecessary big-O complexity when we don't need logging
            setattr(hub, f"log.{log_level}", hub.log.basic.null)



# -*- coding: utf-8 -*-
from typing import Any, Dict, List

try:
    import yaml

    HAS_YAML = True
except ImportError:
    HAS_YAML = False

__virtualname__ = "cli"


def __virtual__(hub):
    if HAS_YAML:
        return True
    return (False, "PyYaml could not be loaded")


def load(hub, path):
    try:
        with open(path, "rb") as fp_:
            ret = {}
            for line in fp_:
                ret.update(hub.config.render.cli.render(line))
            return ret
    except FileNotFoundError:
        pass
    return {}


def render(hub, val: List[str] or str) -> List[str]:
    """
    Take the string and render it in json
    """
    ret = []
    if isinstance(val, str):
        val = [val]
    for v in val:
        if "=" in v:
            key, v = v.split("=", maxsplit=1)
            ret.append({key: yaml.safe_load(v)})
        else:
            ret.append(yaml.safe_load(v))

    return ret
